@@ -101,28 +101,84 @@ class BipartiteLong(bpd.BipartiteLongBase):
         '''
         return self.groupby('i', sort=(not is_sorted))['m'].transform('max').to_numpy() > 0
 
-    def collapse(self, is_sorted=False, copy=True):
+    def clean(self, params=None):
         '''
-        Collapse long data at the worker-firm spell level (so each spell for a particular worker at a particular firm becomes one observation).
+        Clean data to make sure there are no NaN or duplicate observations, observations where workers leave a firm then return to it are removed, firms are connected by movers, and categorical ids are contiguous.
 
         Arguments:
+            params (ParamsDict or None): dictionary of parameters for cleaning. Run bpd.clean_params().describe_all() for descriptions of all valid parameters. None is equivalent to bpd.clean_params().
+
+        Returns:
+            (BipartiteLongBase): dataframe with cleaned data
+        '''
+        if params is None:
+            params = bpd.clean_params()
+        else:
+            params = params.copy()
+
+        self.log('beginning BipartiteLong data cleaning', level='info')
+
+        connectedness = params['connectedness']
+        collapse_connectedness = params['collapse_at_connectedness_measure']
+        # If will have to collapse the data after cleaning
+        collapse = (collapse_connectedness and connectedness in ['leave_out_spell', 'leave_out_match'])
+        if collapse:
+            params['connectedness'] = None
+
+        ## Initial cleaning ##
+        frame = super().clean(params)
+
+        if collapse:
+            ## Collapse then compute largest connected set ##
+            # Update parameters
+            level_dict = {
+                'leave_out_spell': 'spell',
+                'leave_out_match': 'match'
+            }
+            # NOTE: leave-out-observation is equivalent to leave-out-(spell/match) if the data is collapsed at the (spell/match) level, but the code is faster
+            params['connectedness'] = 'leave_out_observation'
+            params['is_sorted'] = True
+            params['copy'] = False
+            params['force'] = False
+
+            # Collapse
+            frame = frame.collapse(level=level_dict[connectedness], is_sorted=True, copy=False)
+
+            # Clean
+            frame = frame.clean(params)
+
+        self.log('BipartiteLongBase data cleaning complete', level='info')
+
+        return frame
+
+    def collapse(self, level='spell', is_sorted=False, copy=True):
+        '''
+        Collapse long data at the worker-firm spell/match level (so each spell/match for a particular worker at a particular firm becomes one observation).
+
+        Arguments:
+            level (str): if 'spell', collapse at the worker-firm spell level; if 'match', collapse at the worker-firm match level ('spell' and 'match' will differ if a worker leaves then returns to a firm)
             is_sorted (bool): if False, dataframe will be sorted by i (and t, if included). Returned dataframe will be sorted. Sorting may alter original dataframe if copy is set to False. Set is_sorted to True if dataframe is already sorted.
             copy (bool): if False, avoid copy
 
         Returns:
             (BipartiteLongCollapsed): collapsed long data generated by collapsing long data at the worker-firm spell level
         '''
-        self.log('beginning collapse', level='info')
+        self.log(f'beginning collapse, level={level!r}', level='info')
 
         # Sort and copy
         frame = self.sort_rows(j_if_no_t=True, is_sorted=is_sorted, copy=copy)
         self.log('data sorted by i (and t, if included; and by j, if t not included)', level='info')
 
-        # Generate spell ids
-        spell_ids = frame._get_spell_ids(is_sorted=True, copy=False)
+        # Generate group ids
+        if level == 'spell':
+            group_ids = frame._get_spell_ids(is_sorted=True, copy=False)
+        elif level == 'match':
+            group_ids = frame.groupby(['i', 'j'], sort=(not frame.no_returns)).ngroup().to_numpy()
+        else:
+            raise ValueError(f"`level` must be one of 'spell' or 'match', but input specifies invalid input {level!r}.")
 
-        if (len(frame) < 2) or (spell_ids[-1] == len(frame)):
-            ## Quickly check whether a collapse is necessary
+        if (len(frame) < 2) or (group_ids[-1] == len(frame)):
+            ## Quickly check whether a collapse is necessary ##
             # Keep track of user-added columns
             user_added_cols = {}
             default_cols = frame.columns_req + frame.columns_opt
@@ -140,9 +196,9 @@ class BipartiteLong(bpd.BipartiteLongBase):
             # Assign data_spell
             data_spell = frame
         else:
-            ### If recollapse necessary
-            ## Aggregate at the spell level
-            spells = frame.groupby(spell_ids, sort=False)
+            ### If recollapse necessary ###
+            ## Aggregate at the spell level ##
+            spells = frame.groupby(group_ids, sort=False)
 
             # Dictionary linking columns to how they should be aggregated
             agg_funcs = {}
@@ -183,7 +239,7 @@ class BipartiteLong(bpd.BipartiteLongBase):
             data_spell = data_spell.reindex(sorted_cols, axis=1, copy=False)
             data_spell.reset_index(drop=True, inplace=True)
 
-        self.log('data aggregated at the spell level', level='info')
+        self.log(f'data aggregated at the {level!r} level', level='info')
 
         collapsed_frame = bpd.BipartiteLongCollapsed(data_spell, col_reference_dict=user_added_cols, log=frame._log_on_indicator)
         collapsed_frame._set_attributes(frame, no_dict=True)
@@ -196,18 +252,18 @@ class BipartiteLong(bpd.BipartiteLongBase):
                 del collapsed_frame.col_collapse_dict[col]
                 del collapsed_frame.col_long_es_dict[col]
                 if col in collapsed_frame.columns_contig.keys():
-                    # If column is contiguous
+                    # If column is categorical
                     del collapsed_frame.columns_contig[col]
                     if collapsed_frame.id_reference_dict:
                         # If linking contiguous ids to original ids
                         del collapsed_frame.id_reference_dict[col]
 
-        if not ((len(frame) < 2) or (spell_ids[-1] == len(frame))):
+        if not ((len(frame) < 2) or (group_ids[-1] == len(frame))):
             # If data actually collapsed, m can change from long to collapsed long
             collapsed_frame = collapsed_frame.gen_m(force=True, copy=False)
 
-        # If no time column, then returns will be collapsed
-        if not collapsed_frame._col_included('t'):
+        # If no time column, or level == 'match', then returns will be collapsed
+        if not collapsed_frame._col_included('t') or level == 'match':
             collapsed_frame.no_returns = True
 
         return collapsed_frame
@@ -268,6 +324,9 @@ class BipartiteLong(bpd.BipartiteLongBase):
         Returns:
             (NumPy Array): indices of articulation observations
         '''
+        if not is_sorted:
+            raise NotImplementedError('._get_articulation_observations() requires `is_sorted` == True, but it is set to False.')
+
         # Find bridges (recall i is adjusted to be greater than j, which is why we reverse the order) (source: https://igraph.discourse.group/t/function-to-find-edges-which-are-bridges-in-r-igraph/154/2)
         bridges = [tuple(sorted(a, reverse=True)) for a in G.biconnected_components() if len(a) == 2]
         bridges_workers = set([bridge[0] - (max_j + 1) for bridge in bridges])
@@ -295,6 +354,9 @@ class BipartiteLong(bpd.BipartiteLongBase):
         Returns:
             (NumPy Array): indices of articulation spells
         '''
+        if not is_sorted:
+            raise NotImplementedError('._get_articulation_spells() requires `is_sorted` == True, but it is set to False.')
+
         # Sort and copy
         frame = self.sort_rows(is_sorted=is_sorted, copy=copy)
 
